// addrspace.cc 
//	Routines to manage address spaces (executing user programs).
//
//	In order to run a user program, you must:
//
//	1. link with the -N -T 0 option 
//	2. run coff2noff to convert the object file to Nachos format
//		(Nachos object code format is essentially just a simpler
//		version of the UNIX executable object code format)
//	3. load the NOFF file into the Nachos file system
//		(if you haven't implemented the file system yet, you
//		don't need to do this last step)
//
// Copyright (c) 1992-1993 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.

#include "copyright.h"
#include "system.h"
#include "addrspace.h"
#include "noff.h"
#ifdef HOST_SPARC
#include <strings.h>
#endif

#include "bitmap.h"
extern BitMap* bitmap_page_in_use;

//----------------------------------------------------------------------
// SwapHeader
// 	Do little endian to big endian conversion on the bytes in the 
//	object file header, in case the file was generated on a little
//	endian machine, and we're now running on a big endian machine.
//----------------------------------------------------------------------

static void 
SwapHeader (NoffHeader *noffH)
{
	noffH->noffMagic = WordToHost(noffH->noffMagic);
	noffH->code.size = WordToHost(noffH->code.size);
	noffH->code.virtualAddr = WordToHost(noffH->code.virtualAddr);
	noffH->code.inFileAddr = WordToHost(noffH->code.inFileAddr);
	noffH->initData.size = WordToHost(noffH->initData.size);
	noffH->initData.virtualAddr = WordToHost(noffH->initData.virtualAddr);
	noffH->initData.inFileAddr = WordToHost(noffH->initData.inFileAddr);
	noffH->uninitData.size = WordToHost(noffH->uninitData.size);
	noffH->uninitData.virtualAddr = WordToHost(noffH->uninitData.virtualAddr);
	noffH->uninitData.inFileAddr = WordToHost(noffH->uninitData.inFileAddr);
}

//----------------------------------------------------------------------
// AddrSpace::AddrSpace
// 	Create an address space to run a user program.
//	Load the program from a file "executable", and set everything
//	up so that we can start executing user instructions.
//
//	Assumes that the object code file is in NOFF format.
//
//	First, set up the translation from program memory to physical 
//	memory.  For now, this is really simple (1:1), since we are
//	only uniprogramming, and we have a single unsegmented page table
//
//	"executable" is the file containing the object code to load into memory
//----------------------------------------------------------------------

AddrSpace::AddrSpace(OpenFile *executable)
{
    NoffHeader noffH;
    unsigned int i, size;

    executable->ReadAt((char *)&noffH, sizeof(noffH), 0);
    if ((noffH.noffMagic != NOFFMAGIC) && 
		(WordToHost(noffH.noffMagic) == NOFFMAGIC))
    	SwapHeader(&noffH);
    ASSERT(noffH.noffMagic == NOFFMAGIC);

// how big is address space?
    size = noffH.code.size + noffH.initData.size + noffH.uninitData.size 
			+ UserStackSize;	// we need to increase the size
						// to leave room for the stack
    numPages = divRoundUp(size, PageSize);
    size = numPages * PageSize;
	//EECE 315 assignment 3, task 1
    //ASSERT(numPages <= bitmap_page_in_use->NumClear());		// check we're not trying
						// to run anything too big --
						// at least until we have
						// virtual memory

    DEBUG('a', "Initializing address space, num pages %d, size %d\n", 
					numPages, size);
// first, set up the translation 
    pageTable = new TranslationEntry[numPages];
    for (i = 0; i < numPages; i++) {
	pageTable[i].virtualPage = i;	// for now, virtual page # = phys page #

	//TODO: EECE 315 Assignment 4
	//For task 2
	//Lazy loading: do not allocate physical page until unless page fault occurs
	//Pagefault exception handling has been added in the exception.cc file
	//you need to call this allocation there
	//Please refer that for futher details
	//Use prepare page function to store data in physical page (Signature defined at the end of this file)
	pageTable[i].physicalPage = bitmap_page_in_use->Find();

	//TODO: EECE 315 Assignment 4
	//For task 1
	//page should only be valid once you allocate the page
	//so you can set valid bit to false.. this will initiate the pagefault exceptions
	pageTable[i].valid = TRUE;

	pageTable[i].use = FALSE;
	pageTable[i].dirty = FALSE;
	pageTable[i].readOnly = FALSE;  // if the code segment was entirely on 
					// a separate page, we could set its 
					// pages to be read-only


	//TODO: EECE 315 Assignment 4
	//For task 2
	//as you are not allocating memory, you need not clear it also here..
	//you can remove this
	bzero(machine->mainMemory + (pageTable[i].physicalPage * PageSize), PageSize);
    }


	//TODO: EECE 315 Assignment 4
	//For task 1 i.e. Demand paging
	//This code and data allocation should be moved preparePage function

	//TODO: EECE 315 Assignment 4
	//For task 2 i.e virtual memory
	//you need to modify this code to remove pages from physical memory when there is no empty page available
	//also you need to add the pages back when page fault occurs
	//you will need to implement and use BackingStore class to store the data stores at physical address being pagedout.
	
		
	i = 0;
	for(; noffH.code.size - PageSize > 0; i++)
	{
		executable->ReadAt(&(machine->mainMemory[(pageTable[i].physicalPage) * PageSize]), PageSize, noffH.code.inFileAddr);
		noffH.code.inFileAddr += PageSize;
		noffH.code.size -= PageSize;
	}
	if(noffH.code.size > 0 && noffH.initData.size > 0)
	{
		executable->ReadAt(&(machine->mainMemory[(pageTable[i].physicalPage) * PageSize]), noffH.code.size, noffH.code.inFileAddr);
		if(PageSize - noffH.code.size > noffH.initData.size)
		{
			DEBUG('a', "Initializing data segment, at 0x%x, size %d\n", 
				noffH.initData.virtualAddr, noffH.initData.size);
			executable->ReadAt(&(machine->mainMemory[(pageTable[i].physicalPage) * PageSize + noffH.code.size]), noffH.initData.size, noffH.initData.inFileAddr);
			noffH.initData.size -= noffH.initData.size;
			noffH.initData.inFileAddr += noffH.initData.size;
		}
		else
		{
			executable->ReadAt(&(machine->mainMemory[(pageTable[i].physicalPage) * PageSize + noffH.code.size]), PageSize - noffH.code.size, noffH.initData.inFileAddr);
			noffH.initData.size -= PageSize - noffH.code.size;
			noffH.initData.inFileAddr += PageSize - noffH.code.size;
		}
		i++;
	}
	for(; noffH.initData.size - PageSize > 0; i++)
	{
		executable->ReadAt(&(machine->mainMemory[(pageTable[i].physicalPage) * PageSize]), PageSize, noffH.initData.inFileAddr);
		noffH.initData.inFileAddr += PageSize;
		noffH.initData.size -= PageSize;
	}
	if(noffH.initData.size > 0)
	{
		executable->ReadAt(&(machine->mainMemory[(pageTable[i].physicalPage) * PageSize]), noffH.initData.size, noffH.initData.inFileAddr);
	}

}

//----------------------------------------------------------------------
// AddrSpace::~AddrSpace
// 	Dealloate an address space.  Nothing for now!
//----------------------------------------------------------------------

AddrSpace::~AddrSpace()
{
	for(int i = 0; i < numPages; i++)
	{
		bitmap_page_in_use->Clear(pageTable[i].physicalPage);
	}
   delete pageTable;
}

//----------------------------------------------------------------------
// AddrSpace::InitRegisters
// 	Set the initial values for the user-level register set.
//
// 	We write these directly into the "machine" registers, so
//	that we can immediately jump to user code.  Note that these
//	will be saved/restored into the currentThread->userRegisters
//	when this thread is context switched out.
//----------------------------------------------------------------------

void
AddrSpace::InitRegisters()
{
    int i;

    for (i = 0; i < NumTotalRegs; i++)
	machine->WriteRegister(i, 0);

    // Initial program counter -- must be location of "Start"
    machine->WriteRegister(PCReg, 0);	

    // Need to also tell MIPS where next instruction is, because
    // of branch delay possibility
    machine->WriteRegister(NextPCReg, 4);

   // Set the stack register to the end of the address space, where we
   // allocated the stack; but subtract off a bit, to make sure we don't
   // accidentally reference off the end!
    machine->WriteRegister(StackReg, numPages * PageSize - 16);
    DEBUG('a', "Initializing stack register to %d\n", numPages * PageSize - 16);
}

//----------------------------------------------------------------------
// AddrSpace::SaveState
// 	On a context switch, save any machine state, specific
//	to this address space, that needs saving.
//
//	For now, nothing!
//----------------------------------------------------------------------

void AddrSpace::SaveState() 
{}

//----------------------------------------------------------------------
// AddrSpace::RestoreState
// 	On a context switch, restore the machine state so that
//	this address space can run.
//
//      For now, tell the machine where to find the page table.
//----------------------------------------------------------------------

void AddrSpace::RestoreState() 
{
    machine->pageTable = pageTable;
    machine->pageTableSize = numPages;
}

void AddrSpace::preparePage(int badVAddr)
{
	//TODO: EECE 315 Assignment 4
	//prepare the content of the request page
	//copy the content from executable or backing store
	//set the valid bit
	
}

void BackingStore::PageOut(void* thread_ptr, int vitual_page, int phy_addr, int size)
{
	//TODO: EECE 315 Assignment 4
	//Use this function to store data from physical address
	// you can either use Nachos File API or declare any data structure for storing
}
void BackingStore::PageIn(void* thread_ptr, int vitual_page, int phy_addr, int size)
{
	//TODO: EECE 315 Assignment 4
	//use this function to read the stored values and restore it at the given physical address
}

